#!/usr/bin/perl

# AGIVS Library for Perl, Version 0.9/beta
# by Nat Budin
# (c) 2000-2001 DB Interactive

# This program is distributed under the terms of the GNU General
# Public License.  For more information, see http://www.gnu.org/.

# This file is part of the AGIVS online version-control system for multi-
# developer AGI games.  To use it in a Perl script, add the following
# command to the top of your code:
#
# require "agivs.inc";
#
# Before calling the functions, you should have the variable $agidir set to
# the path to your AGI game, and the variable $dirseparator set to the
# appropriate directory separator for your OS ("/" for Unix, "\\" for DOS
# or Windows, ":" for MacOS).

use File::Path;

$LOGIC = 1;
$PICTURE = 2;
$SOUND = 3;
$VIEW = 4;
$RAW = 5;
$typename[$LOGIC] = "LOGIC";
$typename[$PICTURE] = "PICTURE";
$typename[$SOUND] = "SOUND";
$typename[$VIEW] = "VIEW";
$typename[$RAW] = "RAW";

# cv (string, offset, length)
# Utility function that returns the value of the packed binary number at
# specified offset in the string, having the specified number of bytes.
# If length is not specified, it is assumed to be 1.
sub cv {
	my ($string, $offset, $length) = @_;
	my ($cntr, $retval);
	$retval = 0;
	if (!$length) {
		$length = 1;
	}
	for ($cntr = 0; $cntr<$length; $cntr++) {
		$retval += ord(substr($string, $offset + $cntr, 1)) * (256 ** $cntr);
	}
	return $retval;
}

# filelen (filehandle)
# Returns the file length of the specified filehandle in bytes.
sub filelen {
	my $FILEHANDLE = $_[0];
	my $savepos = tell($FILEHANDLE);
	seek($FILEHANDLE,0,2);
	my $retval = tell($FILEHANDLE)+1;
	seek($FILEHANDLE,$savepos,0);
	return $retval;
}

# clear_dir (dirarray)
# Fills the specified array with entries of VOL 0xF, offset 0xFFFFF (in
# other words, resource does not exist).
sub clear_dir {
	my $dirarray = $_[0];
	my ($cntr, $dirrec);
	@{ $dirarray } = ();
	for ($cntr = 0; $cntr<256; $cntr++) {
		$dirrec = {};
		$dirrec->{vol} = 0xF;
		$dirrec->{offset} = 0xFFFFF;
		push @{ $dirarray }, $dirrec;
	}
}

# read_dirs (no arguments)
# Sets up the variables necessary to find specific resources by reading
# in the files LOGDIR, PICDIR, SNDDIR, and VIEWDIR.  You must call
# this function before using any of the input functions in this library.
sub read_dirs {
	my ($cntr, $dirrec, $bbuf);
	@logdir = ();
	@picdir = ();
	@snddir = ();
	@viewdir = ();
	open(LOGDIR,$agidir.$dirseparator."LOGDIR");
	binmode LOGDIR;
	read(LOGDIR, $bbuf, filelen(\*LOGDIR));
	for ($cntr=0; $cntr<length($bbuf); $cntr+=3) {
		$dirrec = {};
		$dirrec->{vol} = (ord(substr($bbuf,$cntr,1)) & 0xF0) >> 4;
		$dirrec->{offset} = ((ord(substr($bbuf,$cntr,1)) & 0x0F) * 256 * 256) + ord(substr($bbuf,$cntr+1,1)) * 256 + ord(substr($bbuf,$cntr+2,1));
		push @logdir, $dirrec;
	}
	close LOGDIR;
	if (@logdir < 256) {
		my $numtoadd = 256 - @logdir;
		for (my $ac = 0; $ac<$numtoadd; $ac++) {
			$dirrec = {};
			$dirrec->{vol} = 0xF;
			$dirrec->{offset} = 0xFFFFF;
			push @logdir, $dirrec;
		}
	}
	open(PICDIR,$agidir.$dirseparator."PICDIR"); 
	binmode PICDIR;
	read(PICDIR, $bbuf, filelen(\*PICDIR));
	for ($cntr=0; $cntr<length($bbuf); $cntr+=3) {
		$dirrec = {};
		$dirrec->{vol} = (ord(substr($bbuf,$cntr,1)) & 0xF0) >> 4;
		$dirrec->{offset} = ((ord(substr($bbuf,$cntr,1)) & 0x0F) * 256 * 256) + ord(substr($bbuf,$cntr+1,1)) * 256 + ord(substr($bbuf,$cntr+2,1));
		push @picdir, $dirrec;
	}
	close PICDIR;
	if (@picdir < 256) {
		my $numtoadd = 256 - @picdir;
		for (my $ac = 0; $ac<$numtoadd; $ac++) {
			$dirrec = {};
			$dirrec->{vol} = 0xF;
			$dirrec->{offset} = 0xFFFFF;
			push @picdir, $dirrec;
		}
	}
	open(SNDDIR,$agidir.$dirseparator."SNDDIR");
	binmode SNDDIR;
	read(SNDDIR, $bbuf, filelen(\*SNDDIR));
	for ($cntr=0; $cntr<length($bbuf); $cntr+=3) {
		$dirrec = {};
		$dirrec->{vol} = (ord(substr($bbuf,$cntr,1)) & 0xF0) >> 4;
		$dirrec->{offset} = ((ord(substr($bbuf,$cntr,1)) & 0x0F) * 256 * 256) + ord(substr($bbuf,$cntr+1,1)) * 256 + ord(substr($bbuf,$cntr+2,1));
		push @snddir, $dirrec;
	}
	close SNDDIR;
	if (@snddir < 256) {
		my $numtoadd = 256 - @snddir;
		for (my $ac = 0; $ac<$numtoadd; $ac++) {
			$dirrec = {};
			$dirrec->{vol} = 0xF;
			$dirrec->{offset} = 0xFFFFF;
			push @snddir, $dirrec;
		}
	}
	open(VIEWDIR,$agidir.$dirseparator."VIEWDIR");
	binmode VIEWDIR;
	read(VIEWDIR, $bbuf, filelen(\*VIEWDIR));
	for ($cntr=0; $cntr<length($bbuf); $cntr+=3) {
		$dirrec = {};
		$dirrec->{vol} = (ord(substr($bbuf,$cntr,1)) & 0xF0) >> 4;
		$dirrec->{offset} = ((ord(substr($bbuf,$cntr,1)) & 0x0F) * 256 * 256) + ord(substr($bbuf,$cntr+1,1)) * 256 + ord(substr($bbuf,$cntr+2,1));
		push @viewdir, $dirrec;
	}
	close VIEWDIR;
	if (@viewdir < 256) {
		my $numtoadd = 256 - @viewdir;
		for (my $ac = 0; $ac<$numtoadd; $ac++) {
			$dirrec = {};
			$dirrec->{vol} = 0xF;
			$dirrec->{offset} = 0xFFFFF;
			push @viewdir, $dirrec;
		}
	}
}

# show_dir (dirarray)
# Prints out the resource names and VOL numbers in the specified dir
# array (@logdir, @picdir, @snddir, or @viewdir).
sub show_dir {
	my @dirarray = @_;
	my $cntr;
	for ($cntr=0; $cntr<@dirarray; $cntr++) {
		if ($dirarray[$cntr]{offset} != 0xFFFFF || $dirarray[$cntr]{vol} != 0x0F) {
			print "$cntr: VOL.$dirarray[$cntr]{vol}, offset $dirarray[$cntr]{offset}\n";
		}
	}
}

# vol_get_resource (type, number)
# Returns the specified resource as a scalar string.
sub vol_get_resource {
	my ($type, $number) = @_;
	my (@cdir, $reslen, $retval);
	if ($type == $LOGIC) {
		@cdir = @logdir;
	}
	elsif ($type == $PICTURE) {
		@cdir = @picdir;
	}
	elsif ($type == $SOUND) {
		@cdir = @snddir;
	}
	elsif ($type == $VIEW) {
		@cdir = @viewdir;
	}
	else {
		return ();
	}
	open(VOLFILE, $agidir.$dirseparator."VOL.$cdir[$number]{vol}");
	binmode VOLFILE;
	seek(VOLFILE, $cdir[$number]{offset}, 0);
	read(VOLFILE, $header, 5);
	$reslen = cv($header, 3, 2);
	read(VOLFILE, $retval, $reslen);
	close VOLFILE;
	return $retval;
}

# save_dirs (no arguments)
# Saves the current DIR variables to DIR files.
sub save_dirs {
	my $cntr;
	open(LOGDIR,">".$agidir.$dirseparator."LOGDIR");
	binmode LOGDIR;
	for ($cntr=0; $cntr<=last_exists($LOGIC); $cntr++) {
		print LOGDIR chr(($logdir[$cntr]{vol} * 16) + (($logdir[$cntr]{offset} & 0xF0000) >> 16));
		print LOGDIR chr(($logdir[$cntr]{offset} & 0xFF00) >> 8);
		print LOGDIR chr($logdir[$cntr]{offset} & 0xFF);
	}
	close LOGDIR;
	open(PICDIR,">".$agidir.$dirseparator."PICDIR"); 
	binmode PICDIR;
	for ($cntr=0; $cntr<=last_exists($PICTURE); $cntr++) {
		print PICDIR chr(($picdir[$cntr]{vol} * 16) + (($picdir[$cntr]{offset} & 0xF0000) >> 16));
		print PICDIR chr(($picdir[$cntr]{offset} & 0xFF00) >> 8);
		print PICDIR chr($picdir[$cntr]{offset} & 0xFF);
	}
	close PICDIR;
	open(SNDDIR,">".$agidir.$dirseparator."SNDDIR");
	binmode SNDDIR;
	for ($cntr=0; $cntr<=last_exists($SOUND); $cntr++) {
		print SNDDIR chr(($snddir[$cntr]{vol} * 16) + (($snddir[$cntr]{offset} & 0xF0000) >> 16));
		print SNDDIR chr(($snddir[$cntr]{offset} & 0xFF00) >> 8);
		print SNDDIR chr($snddir[$cntr]{offset} & 0xFF);
	}
	close SNDDIR;
	open(VIEWDIR,">".$agidir.$dirseparator."VIEWDIR");
	binmode VIEWDIR;
	for ($cntr=0; $cntr<=last_exists($VIEW); $cntr++) {
		print VIEWDIR chr(($viewdir[$cntr]{vol} * 16) + (($viewdir[$cntr]{offset} & 0xF0000) >> 16));
		print VIEWDIR chr(($viewdir[$cntr]{offset} & 0xFF00) >> 8);
		print VIEWDIR chr($viewdir[$cntr]{offset} & 0xFF);
	}
	close VIEWDIR;
}

# vol_write_resource (type, number, volnumber, resource, skipdirs)
# Adds the binary resource data passed in into the specified VOL file
# and updates the DIRs accordingly.  If skipdirs is true, then this
# function will not update the DIRs.
sub vol_write_resource {
	my ($rtype,$rnum,$rvol,$rsrc, $skipdirs) = @_;
	
	if ($rtype == $LOGSRC) {
		return arc_write_resource($rtype,$rnum,time,$rsrc);
	}
	
	open(VOLFILE,">>".$agidir.$dirseparator."VOL.$rvol");
	binmode(VOLFILE);
	my $offset = tell(VOLFILE);
	print VOLFILE chr(0x12),chr(0x34);
	print VOLFILE chr($rvol);
	print VOLFILE chr(length($rsrc) & 0x00FF);
	print VOLFILE chr(length($rsrc) >> 8);
	print VOLFILE $rsrc;
	close VOLFILE;
	
	if ($rtype == $LOGIC) {
		$logdir[$rnum]{offset} = $offset;
		$logdir[$rnum]{vol} = $rvol;
	}
	elsif ($rtype == $PICTURE) {
		$picdir[$rnum]{offset} = $offset;
		$picdir[$rnum]{vol} = $rvol;
	}
	elsif ($rtype == $SOUND) {
		$snddir[$rnum]{offset} = $offset;
		$snddir[$rnum]{vol} = $rvol;
	}
	elsif ($rtype == $VIEW) {
		$viewdir[$rnum]{offset} = $offset;
		$viewdir[$rnum]{vol} = $rvol;
	}
		
	unless ($skipdirs) { save_dirs() };
}

# resource_exists (restype, number)
# Returns true if the specified resource exists in the VOL files, or 
# false if it doesn't.
sub resource_exists {
	my ($restype, $number) = @_;
		
	if ($restype == $LOGIC) {
		return (($logdir[$number]{vol} != 0xF) && ($logdir[$number]{offset} != 0xFFFFF));
	}
	elsif ($restype == $PICTURE) {
		return (($picdir[$number]{vol} != 0xF) && ($picdir[$number]{offset} != 0xFFFFF));
	}
	elsif ($restype == $SOUND) {
		return (($snddir[$number]{vol} != 0xF) && ($snddir[$number]{offset} != 0xFFFFF));
	}
	elsif ($restype == $VIEW) {
		return (($viewdir[$number]{vol} != 0xF) && ($viewdir[$number]{offset} != 0xFFFFF));
	}
	
	return 0;
}

# last_exists (restype)
# Returns the number of the last resource of a particular type
# that exists in the VOL files.
sub last_exists {
	my ($restype) = @_;
	my $x = 255;
	
	while (!resource_exists($restype,$x) && $x>0) {
		$x--;
	}
	
	return $x;
}

# arc_get_resource (type, number, timestamp)
# Returns a resource file read from the archives at the specified
# type, number, and timestamp.
sub arc_get_resource {
	my ($name,$timestamp) = @_;
	my $retval;
	$name =~ s/\//./g;
	open (RSRCFILE, $agidir.$dirseparator."arc".$dirseparator."$name.$timestamp");
	binmode (RSRCFILE);
	read (RSRCFILE, $retval, filelen(\*RSRCFILE));
	close (RSRCFILE);
	return $retval;
}

# raw_get_resource (type, number, timestamp)
# Returns a resource file read from a rawfile with the specified
# name.
sub raw_get_resource {
	my ($name) = @_;
	my $retval;
	$name =~ s/\//$dirseparator/g;
	open (RSRCFILE, $agidir.$dirseparator."$name");
	binmode (RSRCFILE);
	read (RSRCFILE, $retval, filelen(\*RSRCFILE));
	close (RSRCFILE);
	return $retval;
}

# arc_write_resource (type, number, timestamp, resource)
# Writes the specified resource to the archives at the specified
# type, number, and timestamp.
sub arc_write_resource {
	my ($name, $timestamp, $rsrc) = @_;
	$name =~ s/\//./g;
	open (RSRCFILE, ">".$agidir.$dirseparator."arc".$dirseparator."$name.$timestamp");
	binmode (RSRCFILE);
	print RSRCFILE $rsrc;
	close RSRCFILE;
}

# raw_write_resource (fn, resource)
# Writes the specified resource to a raw file with the specified
# filename, and timestamp.
sub raw_write_resource {
	my ($name, $rsrc) = @_;
	$name =~ s/\//$dirseparator/g;
	if (!open (RSRCFILE, ">".$agidir.$dirseparator.$name)) {
		my $filepath = $name;
		$filepath =~ s/[\/\\][^\/\\]+$//;
		mkpath($agidir.$dirseparator.$filepath);
		open (RSRCFILE, ">".$agidir.$dirseparator.$name);
	}
	binmode (RSRCFILE);
	print RSRCFILE $rsrc;
	close RSRCFILE;
}

# setup_version_file (no arguments)
# To be used at the beginning of an AGIVS client setup.  Creates
# an "AGIVERS" file in the $agidir directory and sets the timestamp
# on all resources to now, then calls read_version_file.
sub setup_version_file {
	my $cntr;
	my $ctimestamp = time;
	my $arcexists = 0;
	opendir(MAINDIR, $agidir);
	my @maindirfiles = readdir MAINDIR;
	closedir(MAINDIR);
	if (!grep /arc/i, @maindirfiles) {
		mkdir $agidir.$dirseparator."arc", 0777;
	}
	opendir(SRCDIR, $agidir.$dirseparator."src");
	my @srcfiles = grep !/^\./, readdir SRCDIR;
	closedir(SRCDIR);
	open(VERSFILE, ">".$agidir.$dirseparator."AGIVERS");
	foreach $cntr (@srcfiles) {
		print VERSFILE "raw:src/$cntr = $ctimestamp\n";
		arc_write_resource("src/$cntr",$ctimestamp,raw_get_resource("src/$cntr"));
	}
	for ($cntr=0; $cntr<256; $cntr++) {
		if (resource_exists($LOGIC,$cntr)) {
			print VERSFILE "vol:LOGIC.$cntr = $ctimestamp\n";
			arc_write_resource("LOGIC.$cntr",$ctimestamp,vol_get_resource($LOGIC,$cntr));
		}
		if (resource_exists($PICTURE,$cntr)) {
			print VERSFILE "vol:PICTURE.$cntr = $ctimestamp\n";
			arc_write_resource("PICTURE.$cntr",$ctimestamp,vol_get_resource($PICTURE,$cntr));
		}
		if (resource_exists($SOUND,$cntr)) {
			print VERSFILE "vol:SOUND.$cntr = $ctimestamp\n";
			arc_write_resource("SOUND.$cntr",$ctimestamp,vol_get_resource($SOUND,$cntr));
		}
		if (resource_exists($VIEW,$cntr)) {
			print VERSFILE "vol:VIEW.$cntr = $ctimestamp\n";
			arc_write_resource("VIEW.$cntr",$ctimestamp,vol_get_resource($VIEW,$cntr));
		}
	}

	if (grep /words\.tok/i, @maindirfiles) {
		print VERSFILE "raw:WORDS.TOK = $ctimestamp\n";
		arc_write_resource("WORDS.TOK",$ctimestamp,raw_get_resource("WORDS.TOK"));
	}
	if (grep /object/i, @maindirfiles) {
		print VERSFILE "raw:OBJECT = $ctimestamp\n";
		arc_write_resource("OBJECT",$ctimestamp,raw_get_resource("OBJECT"));
	}
	close VERSFILE;
	read_version_file();
}

# read_version_file (no arguments)
# Parses the AGIVERS file into a hash of hashes called %versions.
sub read_version_file {
	my ($line, $verrec, $lversions, $pversions, $sversions, $vversions, $cversions);
	%versions = ();
	open(VERSFILE, $agidir.$dirseparator."AGIVERS");
	while ($line = <VERSFILE>) {
		if ($line =~ /(\w+)\:(\S+)\s*\=\s*(\d+)/i) {
			$verrec = {};
			$verrec->{loc} = $1;
			$verrec->{timestamp} = $3;
			push(@{ $versions{$2} }, $verrec);
		}
	}
}

# read_version_array (versarray)
# Parses an AGIVERS-structured array of lines into an hash of hashes 
# and returns this compound hash.  To be used when synchronizing 
# VOLs with another computer.
sub read_version_array {
	my @versfile = @_;
	my ($line, $verrec, $lversions, $pversions, $sversions, $vversions);
	my %rversions = ();
	foreach $line (@versfile) {
		if ($line =~ /(\w+)\:(\S+)\s*\=\s*(\d+)/i) {
			$verrec = {};
			$verrec->{loc} = $1;
			$verrec->{timestamp} = $3;
			push(@{ $rversions{$2}}, $verrec);
		}
	}
	return %rversions;
}

# write_version_file (no arguments)
# Dumps @volversions and %fileversions into the AGIVERS file.
sub write_version_file {
	my ($type, $number, $version, $fn);
	open(VERSFILE, ">".$agidir.$dirseparator."AGIVERS");
	foreach $fn (keys %versions) {
		foreach $version (@{ $versions{$fn} }) {
			print VERSFILE $version->{loc};
			print VERSFILE ":";
			print VERSFILE $fn;
			print VERSFILE " = ";
			print VERSFILE $version->{timestamp};
			print VERSFILE "\n";
		}
	}
	close VERSFILE;
}

# versions_to_text (no arguments)
# Dumps the @volversions array into a line-by-line array.
sub versions_to_text {
	my ($type, $number, $version, $fn);
	my @verarray = ();
	foreach $fn (keys %versions) {
		foreach $version (@{ $versions{$fn} }) {
			push @verarray, $version->{loc}.":".$fn." = ".$version->{timestamp};
		}
	}
	return @verarray;
}

# update_resource (type, number, resource, loc)
# Adds a resource to the ARCs.  Checks to see if a resource is being
# replaced, and if so, archives the older version.  Updates the
# version files accordingly.
sub update_resource {
	my ($fn, $rsrc, $timestamp, $loc) = @_;
	my ($verrec, $oldver);
	if (!$loc) {
		if ($versions{$fn}) {
			$oldver = latest_version($fn);
			$loc = $oldver->{loc};
		}
		else {
			if ($fn =~ /(LOGIC|VIEW|SOUND|PIC|PICTURE)\.\d{1,3}/i) {
				$loc = "vol";
			}
			else {
				$loc = "arc";
			}
		}
	}
	arc_write_resource($fn,$timestamp,$rsrc);
	$verrec = {};
	$verrec->{loc} = $loc;
	$verrec->{timestamp} = $timestamp;
	push @{ $versions{$fn} }, $verrec;
	write_version_file();
}

# latest_version (fn)
# Returns the most recent version record for the resource you
# specify.
sub latest_version {
	my ($fn) = @_;
	my ($verrec, $latest);
	$latest = {};
	$latest->{timestamp} = 0;
	$latest->{loc} = "none";
	foreach $verrec (@{ $versions{$fn} }) {
		if ($verrec->{timestamp} > $latest->{timestamp}) {
			$latest->{timestamp} = $verrec->{timestamp};
			$latest->{loc} = $verrec->{loc};
		}
	}
	return $latest;
}

# latest_version_v (fn, versarray)
# Returns the most recent version record for the resource you
# specify.
sub latest_version_v {
	my ($fn, %versarray) = @_;
	my ($verrec, $latest);
	$latest = {};
	$latest->{timestamp} = 0;
	$latest->{loc} = "none";
	foreach $verrec (@{ $versarray{$fn} }) {
		if ($verrec->{timestamp} > $latest->{timestamp}) {
			$latest->{timestamp} = $verrec->{timestamp};
			$latest->{loc} = $verrec->{loc};
		}
	}
	return $latest;
}

# latest_resource (type, number)
# Returns the most recent version of the resource you specify.
sub latest_resource {
	my ($fn) = @_;
	my $latest = latest_version($fn);
	return arc_get_resource($fn, $latest->{timestamp});
}

# find_local_updates (versarray)
# Compares a remote versions file to the versions file on your
# machine.  Returns an array of hashes representing the resources
# that are out-of-date on your machine.
sub find_local_updates {
	my @versarray = @_;
	my %rversions = read_version_array(@versarray);
	my @updates = ();
	my $fn;
	foreach $fn (%rversions) {
		my $locallatest = latest_version($fn);
		my $remotelatest = latest_version_v($fn, %rversions);
		if ($locallatest->{timestamp} < $remotelatest->{timestamp}) {
			my $update = {};
			$update->{fn} = $fn;
			$update->{loc} = $remotelatest->{loc};
			$update->{localtimestamp} = $locallatest->{timestamp};
			$update->{remotetimestamp} = $remotelatest->{timestamp};
			push @updates, $update;
		}
	}
	return @updates;
}

# find_remote_updates (versarray)
# Compares a remote versions file to the versions file on your
# machine.  Returns an array of hashes representing the resources
# that are out-of-date on the remote machine.
sub find_remote_updates {
	my @versarray = @_;
	my %rversions = read_version_array(@versarray);
	my @updates = ();
	my $fn;
	foreach $fn (%versions) {
		my $locallatest = latest_version($fn);
		my $remotelatest = latest_version_v($fn, %rversions);
		if ($locallatest->{timestamp} > $remotelatest->{timestamp}) {
			my $update = {};
			$update->{fn} = $fn;
			$update->{loc} = $locallatest->{loc};
			$update->{localtimestamp} = $locallatest->{timestamp};
			$update->{remotetimestamp} = $remotelatest->{timestamp};
			push @updates, $update;
		}
	}
	return @updates;
}

# rebuild_vols (no parameters)
# Rebuilds the VOL files like AGI Studio does, using the version of
# each resource that is already stored in the VOLs.
sub rebuild_vols {
	my @resources = ();
	
	for (my $restype = 1; $restype<5; $restype++) {
		for (my $resnum = 0; $resnum < 256; $resnum++) {
			if (resource_exists($restype,$resnum)) {
				$resources[$restype][$resnum] = vol_get_resource($restype,$resnum);
			}
		}
	}
	
	clear_dir(\@logdir);
	clear_dir(\@snddir);
	clear_dir(\@picdir);
	clear_dir(\@viewdir);
	
	open(VOLFILE,">".$agidir.$dirseparator."VOL.0");
	close(VOLFILE);
	
	for (my $restype = 1; $restype<5; $restype++) {
		for (my $resnum = 0; $resnum < 256; $resnum++) {
			if ($resources[$restype][$resnum]) {
				vol_write_resource($restype,$resnum,0,$resources[$restype][$resnum],1);
			}
		}
	}
	
	save_dirs();
}

# rebuild_latest (no parameters)
# Rebuilds the VOL files like AGI Studio does, using the most recent
# version of each resource
sub rebuild_latest {
	my @resources = ();
	
	for (my $restype = 1; $restype<5; $restype++) {
		for (my $resnum = 0; $resnum < 256; $resnum++) {
			my $rsrc = latest_resource($typename[$restype].".".$resnum);
			if ($rsrc) {
				$resources[$restype][$resnum] = $rsrc;
			}
		}
	}
	
	clear_dir(\@logdir);
	clear_dir(\@snddir);
	clear_dir(\@picdir);
	clear_dir(\@viewdir);
	
	open(VOLFILE,">".$agidir.$dirseparator."VOL.0");
	close(VOLFILE);
	
	for (my $restype = 1; $restype<5; $restype++) {
		for (my $resnum = 0; $resnum < 256; $resnum++) {
			if ($resources[$restype][$resnum]) {
				vol_write_resource($restype,$resnum,0,$resources[$restype][$resnum],1);
			}
		}
	}
	
	save_dirs();
}

# volparse (fn)
# Takes a filename and returns an array consisting of (type, number).
sub volparse {
	my ($fn) = @_;
	if ($fn =~ /LOGIC\.(\d+)/i) {
		return ($LOGIC, $1);
	}
	elsif ($fn =~ /SOUND\.(\d+)/i) {
		return ($SOUND, $1);
	}
	elsif ($fn =~ /PICTURE\.(\d+)/i || $fn =~ /PIC\.(\d+)/i) {
		return ($PICTURE, $1);
	}
	elsif ($fn =~ /VIEW\.(\d+)/i) {
		return ($VIEW, $1);
	}
	return ();
}

return 1;
	