#!/usr/local/bin/perl

use LWP::UserAgent;
use MIME::Base64;
use String::CRC32;
use File::Path;
use   URI;
use   LWP::Protocol::HTTP;
use   URI::HTTP;
use   URI::_foreign;                                        # for perlapp
use   HTTP::Headers::Util;

$versig = "AGIVS Client v0.92/beta";

if (check_initial_setup()) {
	print "$versig initial setup\n\n";
	print "Please enter your name: ";
	$username = <>;
	chomp $username;
	if ($username =~ /\s/) {
		$username =~ s/\s//g;
		print "Note: this will be truncated to $username on the Internet.\n\n";
	}
	else {
		print "\n";
	}
	print "Now you must select a working directory.  This directory will store all the\n";
	print "data for your AGI game, as well as special AGIVS data that will enable\n";
	print "synchronization over the Internet.  In order to use Internet synchronization\n";
	print "features, you MUST make any modifications to your AGI game into this\n";
	print "directory (i.e. this should be the working copy of your game which you open\n";
	print "in AGI Studio).\n\n";
	print "If you are setting up a new AGIVS server, this directory should be the one\n";
	print "your AGI game is currently stored in.  Otherwise, it should be a new, empty\n";
	print "directory.  If you enter the name of a directory that does not exist, it\n";
	print "will automatically be created for you.\n\n";
	print "Enter a working directory: ";
	$agidir = <>;
	chomp $agidir;
	$agidir =~ s/[\\\/]$//;
	if ($agidir =~ /\//) {
		$dirseparator = '/';
	}
	elsif ($agidir =~ /\\/) {	# / scite needs this comment because it is brain-damaged.
		$dirseparator = "\\";
	}
	else {
		$dirseparator = "\\";
	}
	print "\n";
	if (!opendir(TEMPDIR, $agidir)) {
		mkpath($agidir);
	}
	else {
		closedir(TEMPDIR);
	}
	print "Enter the name of your AGI game: ";
	$gamename = <>;
	chomp $gamename;
	$gamename =~ s/\=//g;
	print "\nEnter the URL of your AGIVS server: ";
	$server = <>;
	chomp $server;
	print "\n\nAGIVS can be run in a special menu-based mode which will be an easier-to-use\n";
	print "interface than the normal command-prompt interface.  The menu system is highly\n";
	print "recommended.  Would you like to use it? (Y/n) ";
	$menubased = <>;
	chomp $menubased;
	if (!($menubased =~ /\S/)) {
		$menubased = 'Y';
	}
	while (uc($menubased) ne 'Y' && uc($menubased) ne 'N') {
		print "Please enter Y or N: ";
		$menubased = <>;
		chomp $menubased;
	}
	if (uc($menubased = 'Y')) {
		$menubased = 1;
	}
	else {
		$menubased = 0;
	}
	open(CONFFILE,">agivs.conf");
	print CONFFILE "\$\$USERNAME = $username\n";
	print CONFFILE "\$\$DIRSEPARATOR = $dirseparator\n";
	print CONFFILE "\$\$MENUBASED = $menubased\n";
	print CONFFILE "$gamename = $agidir = $server\n";
	close(CONFFILE);
}

open(CONFFILE,"agivs.conf");
while (<CONFFILE>) {
	chomp;
	if ($_ =~ /^\$\$USERNAME\s*\=\s*(.+)/i) {
		$username = $1;
	}
	elsif ($_ =~ /^\$\$DIRSEPARATOR\s*\=\s*(.)/i) {
		$dirseparator = $1;
	}
	elsif ($_ =~ /^\$\$MENUBASED\s*\=\s*(\d)/i) {
		$menubased = $1;
	}
	elsif ($_ =~ /^(.+)?\s+\=\s*(.+)?\s+\=\s+(.+)/i) {
		my $gamerec = {};
		$gamerec->{name} = $1;
		$gamerec->{agidir} = $2;
		$gamerec->{server} = $3;
		push (@choices, $gamerec);
	}
}
close(CONFFILE);

mainmenu:

print "\n$versig main menu\n\n";
for (my $x=0; $x<@choices; $x++) {
	print "  [",$x+1,"] ",$choices[$x]->{name},"\n";
}
print "  [A] Add a new game\n";
print "  [Q] Quit\n\n";
getchoice:
print "Please enter a choice from the menu above: ";
$uch = <>;
chomp $uch;
if (uc($uch) eq "A") {
	print "\nPlease enter the name of the game: ";
	my $gamename = <>;
	chomp $gamename;
	print "Please enter the game's directory: ";
	my $gamedir = <>;
	chomp $gamedir;
	if (!opendir(TEMPDIR, $gamedir)) {
		mkpath($gamedir);
	}
	else {
		closedir(TEMPDIR);
	}
	print "Please enter the URL of the server: ";
	my $gameserver = <>;
	chomp $gameserver;
	my $gamerec = {};
	$gamerec->{name} = $gamename;
	$gamerec->{agidir} = $gamedir;
	$gamerec->{server} = $gameserver;
	push (@choices, $gamerec);
	open(CONFFILE, ">>agivs.conf");
	print CONFFILE "$gamename = $gamedir = $gameserver\n";
	close(CONFFILE);
	goto mainmenu;
}
elsif (uc($uch) eq "Q") {
	print "\nBye.\n";
	exit;
}
elsif ($uch >= 1 && $uch <= @choices) {
	my $gamerec = $choices[$uch-1];
	$agidir = $gamerec->{agidir};
	$server = $gamerec->{server};
}
else {
	goto getchoice;
}

require "agivs.inc";

read_dirs();
if (check_first_time()) {
	print "First time run.  Setting up AGIVERS file...\n";
	setup_version_file();
}
else {
	read_version_file();
	print "Checking for ARC updates";
	my $updates = check_local_arcs(\&print_fn);
	print "\n$updates resources updated.\n\n";
}

$ua = new LWP::UserAgent;
$ua->agent("AGIVS-Client/0.2".$ua->agent);

if ($menubased) {
	@resp = send_command("LOCK $username");
	process_response(@resp);
	print "\nSynchronizing game using $server...\n\n";
	syncgame();
	print "100%\n";
	@resp = send_command("UNLOCK");
	process_response(@resp);
	goto mainmenu;
}

print "$versig shell initialized\n";
print "Using server URL '$server'\n";
print "Enter QUIT to quit\n\n";
print "> ";
$cmd = <>;
chomp $cmd;
while (!($cmd =~ /^quit/i)) {
	if ($cmd =~ /^sync/i) {
		syncgame();
	}
	elsif ($cmd =~ /^rebuild\s*latest/i) {
		rebuild_latest();
	}
	elsif ($cmd =~ /^rebuild/i) {
		rebuild_vols();
	}
	else {
		@resp = send_command($cmd);

		process_response(@resp);
	}
	
	print "\n> ";
	$cmd = <>;
	chomp $cmd;
}

@resp = send_command("UNLOCK");

goto mainmenu;

sub syncgame {
	@resp = send_command("GETVERS");
	if (process_response(@resp)) {
		my @localupdates = find_local_updates(@rversions);
		my @remoteupdates = find_remote_updates(@rversions);
		my $lucount = @localupdates;
		my $rucount = @remoteupdates;
		print "$lucount resources out of date on your machine\n";
		print "$rucount resources out of date on the server\n";
		if ($lucount) {
			print "Downloading updates from the server...\n";
			my $cupd = 0;
			foreach $lupdate (@localupdates) {
				if (int(($cupd/$lucount) * 10) > int((($cupd-1)/$lucount) * 10)) {
					print int(($cupd/$lucount) * 10), "0%  ";
				}
				my $sendcmd = "GETRESOURCE ".$lupdate->{fn}." ".$lupdate->{remotetimestamp};
				@resp = send_command($sendcmd);
				process_response(@resp);
				if ($wroteresource) {
					if ($resloc eq "vol") {
						vol_write_resource(volparse($resfn),0,arc_get_resource($resfn,$rests));
					}
					elsif ($resloc eq "raw") {
						raw_write_resource($resfn,arc_get_resource($resfn,$rests));
					}
				}
				$cupd++;
			}
		}
		if ($rucount) {
			print "Uploading updates to the server...\n";
			my $cupd = 0;
			foreach $rupdate (@remoteupdates) {
				my $resdata;
				if (int(($cupd/$rucount) * 10) > int((($cupd-1)/$rucount) * 10)) {
					print int(($cupd/$rucount) * 10), "0%  ";
				}
				$resdata = arc_get_resource($rupdate->{fn},$rupdate->{localtimestamp});
				my $rawdata = $resdata;
				$resdata = encode_base64($resdata, "\n");
				my @resarry = split(/\n/, $resdata);
				my $linecount = @resarry;
				my $sendcmd = "SENDRESOURCE ".$rupdate->{fn}." ".$linecount." ".$rupdate->{localtimestamp}." ".checksum($rawdata)." ".$rupdate->{loc}."\n".$resdata;
				@resp = send_command($sendcmd);
				process_response(@resp);
				$cupd++;
			}
		}
	}
}

sub send_command {
	my ($cmd) = @_;
	my (@return, @lines, $x);
	my $req = new HTTP::Request(POST => $server);
	$req->content_type('application/agivs-command');
	$req->content($cmd);
	
	if ($agivs_debug) {
		print "\t[$cmd]\n";
	}

	my $res = $ua->request($req);

	if (!$res->is_success) {
		print "Error executing HTTP request\n";
		return ();
	}
	
	@lines = split(/\n/,$res->content);
	
	for ($x=0; $x<@lines; $x++) {
		if ($agivs_debug) {
			print "\t> $lines[$x]\n";
		}
		if ($lines[$x] =~ /TEXT\s*(\d+)/i) {
			$x++;
			my $llen = $1;
			my $rtext = "";
			my $retel = {};
			my $lastline = $x+$llen;
			for ($x=$x; $x<$lastline; $x++) {
				$rtext = $rtext.$lines[$x]."\n";
			}
			$x--;
			$retel->{type} = "TEXT";
			$retel->{value} = $rtext;
			push @return, $retel;
		}
		elsif ($lines[$x] =~ /VERS\s*(\d+)/i) {
			$x++;
			my $llen = $1;
			my $rtext = "";
			my $retel = {};
			my $lastline = $x+$llen;
			for ($x=$x; $x<$lastline; $x++) {
				$rtext = $rtext.$lines[$x]."\n";
			}
			if ($agivs_debug) {
				print "\t> Version data consisting of $llen lines\n";
			}
			$x--;
			$retel->{type} = "VERS";
			$retel->{value} = $rtext;
			push @return, $retel;
		}
		elsif ($lines[$x] =~ /RSRC\s*(\S+)\s*(\d+)\s*(\d+)\s*(\d+)\s*(\w+)/i) {
			$x++;
			my $llen = $2;
			my $rtext = "";
			my $retel = {};
			my $lastline = $x+$llen;
			for ($x = $x; $x<$lastline; $x++) {
				$rtext = $rtext.$lines[$x]."\n";
			}
			$rtext = decode_base64($rtext);
			if ($agivs_debug) {
				print "\t> Resource data consisting of $llen lines, ",length($rtext)," bytes\n";
			}
			$x--;
			$retel->{type} = "RSRC";
			$retel->{value} = $rtext;
			$retel->{fn} = $1;
			$retel->{timestamp} = $3;
			$retel->{checksum} = $4;
			$retel->{loc} = $5;
			push @return, $retel;
		}
		elsif (uc($lines[$x]) eq "ERROR") {
			$x++;
			my $retel = {};
			$retel->{type} = "ERROR";
			$retel->{value} = $lines[$x]."\n";
			push @return, $retel;
		}
		elsif (uc($lines[$x]) eq "END") {
			return @return;
		}
		else {
			print "Server sent back an invalid AGIVS command: $lines[$x]\n";
			return @return;
		}
	}
	
	return @return;
}

sub process_response {
	my @resp = @_;
	my $retval = 1;
	$wroteresource = 0;
	foreach $retel (@resp) {
		if ($retel->{type} eq "TEXT") {
			print $retel->{value};
		}
		elsif ($retel->{type} eq "ERROR") {
			print "Error returned from server:\n   ",$retel->{value};
			$retval = 0;
		}
		elsif ($retel->{type} eq "VERS") {
			print "Versions info received from server.\n";
			@rversions = split(/\n/,$retel->{value});
		}
		elsif ($retel->{type} eq "RSRC") {
			if (checksum($retel->{value}) != $retel->{checksum}) {
				print "Error receiving file ",$retel->{fn},": checksum does not match\n";
				$retval = 0;
			}
			else {
				update_resource($retel->{fn}, $retel->{value}, $retel->{timestamp}, $retel->{loc});
				$wroteresource = 1;
				$resfn = $retel->{fn};
				$rests = $retel->{timestamp};
				$resloc = $retel->{loc};
			}
		}
	}
	return $retval;
}

sub checksum {
	my ($data) = @_;
	my $csum = 0;
	for (my $x=0; $x<length($data); $x++) {
		$csum += ord(substr($data,$x,1));
	}
	return $csum;
}

sub checkrsrc {
	my ($data1, $data2) = @_;
	return (length($data1) != length($data2) || crc32($data1) != crc32($data2));
}

sub checknewrsrc {
	my ($fn) = @_;
	my $latest = latest_version($fn);
	if (!$latest) {
		return 1;
	}
	if ($latest->{loc} eq "vol") {
		my $rsrc = vol_get_resource(volparse($fn));
		if (checkrsrc(arc_get_resource($fn,$latest->{timestamp}), $rsrc)) {
			return ($rsrc, "vol");
		}
		else {
			return 0;
		}
	}
	elsif ($latest->{loc} eq "raw") {
		my $rsrc = raw_get_resource($fn);
		if (checkrsrc(arc_get_resource($fn,$latest->{timestamp}),$rsrc)) {
			return ($rsrc, "vol");
		}
		else {
			return 0;
		}
	}
	return 0;
}

sub check_local_arcs {
	my ($callback) = @_;
	my @filenames = keys(%versions);
	my $timestamp = time;
	my $updates = 0;
	for (my $x=0; $x<@filenames; $x++) {
		my ($rsrc, $loc) = checknewrsrc($filenames[$x]);
		if ($rsrc) {
			update_resource($filenames[$x],$rsrc,$timestamp,$loc);
			$updates++;
		}
		&$callback($x+1,$#filenames+1);
	}
	opendir(SRCDIR,$agidir.$dirseparator."src");
	while ($dirent = readdir(SRCDIR)) {
		if (!$versions{"src/$dirent"}) {
			update_resource("src/$dirent",raw_get_resource("src/$dirent"),$timestamp,"raw");
				$updates++;
		}
	}
	return $updates;
}

sub print_fn() {
	my ($current, $max) = @_;
	if (int(($current/$max) * 10) > int(($current-1)/$max) * 10) {
		print ".";
	}
}

sub check_first_time {
	opendir(AGIDIR,$agidir);
	my $cdir;
	while ($cdir = readdir(AGIDIR)) {
		if ($cdir =~ /^AGIVERS/i) {
			closedir(AGIDIR);
			return 0;
		}
	}
	closedir(AGIDIR);
	return 1;
}

sub check_initial_setup {
	opendir(AGIVSDIR,".");
	my $cdir;
	while ($cdir = readdir(AGIVSDIR)) {
		if ($cdir =~ /^agivs\.conf/i) {
			closedir(AGIVSDIR);
			return 0;
		}
	}
	closedir(AGIDIR);
	return 1;
}