#!/usr/local/bin/perl

use MIME::Base64;

$srvname = "Budin.net OPERATION RECON server";
$lockfile = "agilock";
$versig = "AGIVS Server v0.21/beta";

$agidir = "files";
$dirseparator = "/";

require "agivs.inc";
read_dirs();
read_version_file();

if ($ENV{CONTENT_TYPE} ne 'application/agivs-command') {
	print "Content-type: text/plain\n\n";
	print "This script can only be accessed using an AGIVS client.  Sorry!\n";
	print "[Your browser said: Content-type: ",$ENV{CONTENT_TYPE},"]\n";
	exit;
}

print "Content-type: text/plain\n\n";

if ($ENV{USER_AGENT} =~ /AGIVS-Client\/0.1/) {
	print "TEXT 2\n";
	print "Sorry, you are using a version 0.1 client.  Due to protocol\n";
	print "changes, this server only supports AGIVS version 0.2 and up.\n";
	exit;
}

read(STDIN, $cmd, $ENV{'CONTENT_LENGTH'});

if (uc($cmd) eq "HELLO") {
	# HELLO command
	# No parameters
	# Debugging command to check the version of the server
	# that is running.
	print "TEXT 1\n";
        print "Hello from $versig!\n";
}
elsif (uc($cmd) eq "SRVINFO") {
	# SRVINFO command
	# No parameters
	# Returns the name of this server.
	print "TEXT 1\n";
	print "$srvname\n";
}
elsif ($cmd =~ /LOCK\s*(\w+)/i) {
	# LOCK command
	# Parameter: name
	# Locks the files using the client's IP and the name passed
	# to the server.
	$rval = lockfiles($1);
	if (!$rval) {
		print "TEXT 2\n";
		print "Hello $1\n";
		print "You now have a lock on the game files.\n";
	}
	else {
		print "TEXT 1\n";
		print "Sorry, $rval has a lock on the files right now.\n";
	}
}
elsif ($cmd =~ /UNLOCK/i) {
	# UNLOCK command
	# No parameters
	# Checks to see if the client's IP matches the IP of the
	# client that has locked the files, and if it does, unlocks
	# the files.
	$rval = unlockfiles();
	if ($rval == 0) {
		print "TEXT 1\n";
		print "The files are now unlocked.\n";
	}
	elsif ($rval == -1) {
		print "TEXT 1\n";
		print "The files are already unlocked.\n";
	}
	else {
		print "TEXT 1\n";
		print "Sorry, $rval has a lock on the files right now.\n";
	}
}
elsif ($cmd =~ /GETVERS/i) {
	# GETVERS command
	# No parameters
	# Retrieves the AGIVERS file and sends it to the client.
	my $lock = getlock();
	if ($lock != -1 && $lock != 0) {
		print "ERROR\n";
		print "Sorry, $lock has a lock on the files right now.\n";
	}
	else {
		if ($lock == -1) {
			print "ERROR\n";
			print "You need to lock the files before performing this task.\n";
		}
		else {
			my @versinfo = versions_to_text();
			my $vcount = @versinfo;
			print "VERS $vcount\n";
			foreach $line (@versinfo) {
				print "$line\n";
			}
		}
	}
}
elsif ($cmd =~ /^SENDRESOURCE\s+(\S+)\s+(\d+)\s+(\d+)\s+(\d+)\s+(\w+)/i) {
	# SENDRESOURCE command
	# Parameters: type, number, datalength, timestamp, checksum
	# Receives a resource sent from the client and stores it in the VOL files,
	# updating the AGIVERS file and doing checksum tests to make sure the
	# file came through okay.
	my $fn = $1;
	my $datalen = $2;
	my $timestamp = $3;
	my $checksum = $4;
	my $loc = $5;
	my $lock = getlock();
	if ($lock != -1 && $lock != 0) {
		print "ERROR\n";
		print "Sorry, $lock has a lock on the files right now.\n";
	}
	else {
		if ($lock == -1) {
			print "ERROR\n";
			print "You need to lock the files before performing this task.\n";
		}
		else {
			my @cmdlines = split(/\n/,$cmd);
			my $sentdata = "";
			for (my $x=1; $x<=$datalen; $x++) {
				$sentdata .= "$cmdlines[$x]\n";
			}
			$sentdata = decode_base64($sentdata);
			if (checksum($sentdata) == $checksum) {
				update_resource($fn, $sentdata, $timestamp, $loc);
			}
			else {
				print "ERROR\n";
				print "Checksum does not match!\n";
			}
		}
	}
}
elsif ($cmd =~ /^GETRESOURCE\s+(\S+)\s+(\d+)/i) {
	# GETRESOURCE command
	# Parameters: fn, timestamp
	# Sends a resource to the client.
	my $fn = $1;
	my $timestamp = $2;
	my $lock = getlock();
	if ($lock != -1 && $lock != 0) {
		print "ERROR\n";
		print "Sorry, $lock has a lock on the files right now.\n";
	}
	else {
		if ($lock == -1) {
			print "ERROR\n";
			print "You need to lock the files before performing this task.\n";
		}
		else {
			my $rawdata = arc_get_resource($fn, $timestamp);
			my $resdata = encode_base64($rawdata, "\n");
			my @resarry = split(/\n/, $resdata);
			my $linecount = @resarry;
			my $loc;
			foreach $verrec (@{ $versions{$fn} }) {
				if ($verrec->{timestamp} = $timestamp) {
					$loc = $verrec->{loc};
				}
			}
			print "RSRC ".$fn." ".$linecount." ".$timestamp." ".checksum($rawdata)." ".$loc."\n".$resdata;
		}
	}
}
else {
	print "ERROR\nInvalid AGIVS command: $cmd\n";
}

print "END\n";

exit;

sub lockfiles {
	my ($name) = @_;
	my $lname;
	
	open (LOCKFILE, $lockfile);
	$lname = <LOCKFILE>;
	chomp $lname;
	close (LOCKFILE);
	
	if ($lname =~ /\S/) {
		return $lname;
	}
	else {
		open (LOCKFILE, ">$lockfile");
		print LOCKFILE $name, "\n";
		print LOCKFILE $ENV{REMOTE_ADDR};
		close LOCKFILE;
		return 0;
	}
}

sub unlockfiles {
	my $lname;
	my $laddr;
	
	open (LOCKFILE, $lockfile);
	$lname = <LOCKFILE>;
	$laddr = <LOCKFILE>;
	chomp $lname;
	chomp $laddr;
	close (LOCKFILE);
	
	if ($lname =~ /\S/) {
		if ($laddr ne $ENV{REMOTE_ADDR}) {
			return $lname;
		}
		else {
			open (LOCKFILE, ">$lockfile");
			print LOCKFILE " ";
			close LOCKFILE;
			return 0;
		}
	}
	else {
		return -1;
	}
}

sub getlock {
	my $lname;
	my $laddr;
	
	open (LOCKFILE, $lockfile);
	$lname = <LOCKFILE>;
	$laddr = <LOCKFILE>;
	chomp $lname;
	chomp $laddr;
	close (LOCKFILE);
	
	if ($lname =~ /\S/) {
		if ($laddr ne $ENV{REMOTE_ADDR}) {
			return $lname;
		}
		else {
			return 0;
		}
	}
	else {
		return -1;
	}
}

sub checksum {
	my ($data) = @_;
	my $csum = 0;
	for (my $x=0; $x<length($data); $x++) {
		$csum += ord(substr($data,$x,1));
	}
	return $csum;
}




